use std::fmt;

pub struct Scanner<'a> {
    start: usize,
    current: usize,
    line: u16,
    source: &'a str,
}

impl<'a> fmt::Display for Scanner<'a> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "{}:{}..{} '{}'",
            self.line,
            self.start,
            self.current,
            &self.source[self.start..self.current]
        )
    }
}

impl<'a> Scanner<'a> {
    pub fn new(source: &'a str) -> Self {
        Scanner {
            start: 0,
            current: 0,
            line: 1,
            source,
        }
    }
    pub fn scan_token(&mut self) -> Token {
        self.skip_whitespace();
        self.start = self.current;
        if self.is_at_end() {
            return self.make_token(TokenType::TokenEof);
        }
        let c: char = self.advance();

        match c {
            c if c.is_digit(10) => return self.make_number(),
            c if c.is_alphabetic() || c == '_' => return self.make_identifier(),
            '(' => return self.make_token(TokenType::TokenLeftParen),
            ')' => return self.make_token(TokenType::TokenRightParen),
            '{' => return self.make_token(TokenType::TokenLeftBrace),
            '}' => return self.make_token(TokenType::TokenRightBrace),
            ';' => return self.make_token(TokenType::TokenSemicolon),
            ',' => return self.make_token(TokenType::TokenComma),
            '.' => return self.make_token(TokenType::TokenDot),
            '-' => return self.make_token(TokenType::TokenMinus),
            '+' => return self.make_token(TokenType::TokenPlus),
            '/' => return self.make_token(TokenType::TokenSlash),
            '*' => return self.make_token(TokenType::TokenStar),
            '!' => {
                if self.matches('=') {
                    return self.make_token(TokenType::TokenBangEqual);
                } else {
                    return self.make_token(TokenType::TokenBang);
                }
            }
            '=' => {
                if self.matches('=') {
                    return self.make_token(TokenType::TokenEqualEqual);
                } else {
                    return self.make_token(TokenType::TokenEqual);
                }
            }
            '<' => {
                if self.matches('=') {
                    return self.make_token(TokenType::TokenLessEqual);
                } else {
                    return self.make_token(TokenType::TokenLess);
                }
            }
            '>' => {
                if self.matches('=') {
                    return self.make_token(TokenType::TokenGreaterEqual);
                } else {
                    return self.make_token(TokenType::TokenGreater);
                }
            }
            '"' => return self.make_string(),
            _ => return self.make_error_token("Unexpected character."),
        };
    }
    fn make_identifier(&mut self) -> Token {
        while let Some(c) = self.peek() {
            if c.is_alphabetic() || c == '_' || c.is_digit(10) {
                self.advance();
            } else {
                break;
            }
        }
        return self.make_token(self.identifier_type());
    }
    fn identifier_type(&self) -> TokenType {
        let c = self.source.chars().nth(self.start);

        match c {
            Some('a') => return self.check_keyword(1, 2, "nd", TokenType::TokenAnd),
            Some('c') if self.current - self.start > 1 => {
                match self.source.chars().nth(self.start + 1) {
                    Some('l') => return self.check_keyword(2, 3, "ass", TokenType::TokenClass),
                    Some('a') => return self.check_keyword(2, 1, "p", TokenType::TokenFalse),
                    _ => panic!("bad keyword"),
                }
            }
            Some('e') => return self.check_keyword(1, 3, "lse", TokenType::TokenElse),
            Some('i') => return self.check_keyword(1, 1, "f", TokenType::TokenIf),
            Some('n') if self.current - self.start > 1 => {
                match self.source.chars().nth(self.start + 1) {
                    Some('i') => return self.check_keyword(2, 1, "l", TokenType::TokenNil),
                    Some('o') => return self.check_keyword(2, 0, "", TokenType::TokenFalse),
                    _ => panic!("bad keyword"),
                }
            }
            Some('o') => return self.check_keyword(1, 1, "r", TokenType::TokenOr),
            Some('p') => return self.check_keyword(1, 4, "rint", TokenType::TokenPrint),
            Some('r') => return self.check_keyword(1, 5, "eturn", TokenType::TokenReturn),
            Some('s') => return self.check_keyword(1, 4, "uper", TokenType::TokenSuper),
            Some('v') => return self.check_keyword(1, 2, "ar", TokenType::TokenVar),
            Some('w') => return self.check_keyword(1, 4, "hile", TokenType::TokenWhile),
            Some('f') if self.current - self.start > 1 => {
                match self.source.chars().nth(self.start + 1) {
                    Some('a') => return self.check_keyword(2, 3, "lse", TokenType::TokenFalse),
                    Some('o') => return self.check_keyword(2, 1, "r", TokenType::TokenFor),
                    Some('u') => return self.check_keyword(2, 1, "n", TokenType::TokenFun),
                    _ => panic!("bad keyword"),
                }
            }
            Some('t') if self.current - self.start > 1 => {
                match self.source.chars().nth(self.start + 1) {
                    Some('h') => return self.check_keyword(2, 2, "is", TokenType::TokenThis),
                    Some('r') => return self.check_keyword(2, 2, "ue", TokenType::TokenTrue),
                    _ => panic!("bad keyword"),
                }
            }
            _ => return TokenType::TokenIdentifier,
        };
    }
    fn check_keyword(
        &self,
        start: usize,
        length: usize,
        rest: &str,
        token_type: TokenType,
    ) -> TokenType {
        let end = self.start + start + length;
        let s = &self.source[self.start + start..end];
        let next = self.source.chars().nth(end);
        if (self.current == end) && (s == rest) {
            match next {
                Some(n) if n.is_whitespace() => {
                    return token_type;
                }
                _ => return TokenType::TokenIdentifier,
            }
        } else {
            return TokenType::TokenIdentifier;
        }
    }
    fn make_string(&mut self) -> Token {
        loop {
            match self.peek() {
                Some(c) => match c {
                    '\n' => {
                        self.line += 1;
                        continue;
                    }
                    '"' => {
                        self.advance();
                        return self.make_token(TokenType::TokenString);
                    }
                    _ => {
                        self.advance();
                        continue;
                    }
                },
                None => return self.make_error_token("Unterminated string."),
            }
        }
    }
    fn make_number(&mut self) -> Token {
        while let Some(c) = self.peek() {
            if c.is_digit(10) {
                self.advance();
            } else {
                break;
            }
        }
        match (self.peek(), self.peek_next()) {
            (Some(c), Some(cc)) if c == '.' && cc.is_digit(10) => {
                self.advance();
                while let Some(c) = self.peek() {
                    if c.is_digit(10) {
                        self.advance();
                    } else {
                        break;
                    }
                }
                return self.make_token(TokenType::TokenNumber);
            }
            _ => {
                return self.make_token(TokenType::TokenNumber);
            }
        }
    }
    fn peek(&self) -> Option<char> {
        self.source.chars().nth(self.current)
    }
    fn peek_next(&self) -> Option<char> {
        if self.is_at_end() {
            return None;
        }
        return self.source.chars().nth(self.current + 1);
    }
    fn skip_whitespace(&mut self) {
        loop {
            let peek = self.peek();
            match peek {
                None => return,
                Some(c) => match c {
                    ' ' | '\r' | '\t' => {
                        self.advance();
                        continue;
                    }
                    '\n' => {
                        self.line += 1;
                        self.advance();
                        continue;
                    }
                    '/' => match self.peek_next() {
                        Some(c) if c == '/' => {
                            while let Some(_) = self.peek() {
                                if !self.is_at_end() {
                                    self.advance();
                                }
                            }
                        }
                        _ => return,
                    },
                    _ => return,
                },
            }
        }
    }
    fn is_at_end(&self) -> bool {
        return self.current == self.source.len();
    }
    fn make_token(&self, token_type: TokenType) -> Token {
        Token {
            token_type,
            start: self.start,
            lexeme: self.source[self.start..self.current].to_string(),
            line: self.line,
        }
    }
    fn advance(&mut self) -> char {
        self.current += 1;

        match self.source.chars().nth(self.current - 1) {
            Some(c) => c,
            None => panic!(
                "advance failed: fell off the end. {}:{}",
                self.line, self.current
            ),
        }
    }
    fn matches(&mut self, expected: char) -> bool {
        match self.source.chars().nth(self.current) {
            Some(c) => {
                if c != expected {
                    return false;
                }
                self.current += 1;
                return true;
            }
            None => return false,
        }
    }
    fn make_error_token(&self, message: &str) -> Token {
        Token {
            token_type: TokenType::TokenError,
            start: self.start,
            lexeme: message.to_string(),
            line: self.line,
        }
    }
}

pub struct Token {
    pub token_type: TokenType,
    pub start: usize,
    pub lexeme: String,
    pub line: u16,
}

impl Token {}

#[derive(Clone, Copy, strum_macros::Display)]
pub enum TokenType {
    // Single-character tokens.
    TokenLeftParen,
    TokenRightParen,
    TokenLeftBrace,
    TokenRightBrace,
    TokenComma,
    TokenDot,
    TokenMinus,
    TokenPlus,
    TokenSemicolon,
    TokenSlash,
    TokenStar,
    // One or two character tokens.
    TokenBang,
    TokenBangEqual,
    TokenEqual,
    TokenEqualEqual,
    TokenGreater,
    TokenGreaterEqual,
    TokenLess,
    TokenLessEqual,
    // Literals.
    TokenIdentifier,
    TokenString,
    TokenNumber,
    // Keywords.
    TokenAnd,
    TokenClass,
    TokenElse,
    TokenFalse,
    TokenFor,
    TokenFun,
    TokenIf,
    TokenNil,
    TokenOr,
    TokenPrint,
    TokenReturn,
    TokenSuper,
    TokenThis,
    TokenTrue,
    TokenVar,
    TokenWhile,

    TokenError,
    TokenEof,
}
